import java.util.*;
import java.util.concurrent.*;
import java.util.concurrent.atomic.*;

import java.util.Random;

public class ListTest implements ListDriver.SBenchmark
{
    private ListSet l;
    static final int MAXNUM = 1024;
    private intR nums[];

    public class intR
    {
	int n;
	intR(int n) {this.n = n;}
	int get() {return n;}
	void set(int n) {this.n = n;}
    }

    public Thread createTestThread(
	String sClassName, ListCounter counter, RWMix mix)
    {
	try {
	    if (null == l) {
		Class sClass = Class.forName(sClassName);
		l = (ListSet)sClass.newInstance();
		nums = new intR[MAXNUM];
		for (int i = 0; i < MAXNUM; i++)
		    nums[i] = new intR(i);
	    }
	} catch (Exception e) {
            e.printStackTrace(System.out);
	}

	return new TestThread(l, nums, MAXNUM, counter, mix);
    }

    public Thread createLoadThread(ListCounter ctr)
    {
	return new LoadThread(ctr);
    }

    public void initializeTestEnvironment(int n)
    {
	Random random = new Random();
	int i = 0;
	while (i < n)
	{
	    int v = random.nextInt();
	    v = v & (MAXNUM - 1);
            if (l.add(nums[v]) > 0)
		i++;
	}
    }

    class LoadThread extends java.lang.Thread 
    {
	ListCounter ctr;
	volatile int i, j;
	LoadThread(ListCounter counter)
	{
	    ctr = counter;
	}
    
	public void run()
	{
	    while (!ctr.timeToStop())
	    {
		// Do some meaningless computation
		i += j++;
		j *= i;
	    }
	}
    }

    class TestThread extends java.lang.Thread 
    {
	ListCounter ctr;
	ListSet l;
	intR nums[];
	int r,a,d;
	Random random;
	int maxnum;
	int randomList[];
	int randIdx;
	final int randomListSize = 1024;

	TestThread(ListSet l, intR nums[], int max, ListCounter ctr, RWMix mix)
	{
	    this.l = l;
	    this.nums = nums;
	    this.ctr = ctr;
	    r = mix.readPct;
	    a = mix.addPct;
	    d = mix.delPct;
	    random = new Random();
	    randomList = new int[randomListSize];
	    for (randIdx = 0; randIdx < randomListSize; randIdx++)
		randomList[randIdx] = random.nextInt();
	    maxnum = max;
	}

	int nextRand() {
	    return randomList[++randIdx % randomListSize];
	}
    
	public void run()
	{
	    int reads = 0;
	    int adds = 0;
	    int deletes = 0;
	    int successes = 0;
	    int failures = 0;

	    int result;

	    while (!ctr.timeToStop())
	    {
		int op = nextRand() % 100;
		if (op < 0) op *= -1;
		int val = nextRand() % maxnum;
		if (val < 0) val *= -1;

		if (op < r) {
		    if (l.contains(nums[val]))
			successes++;
		    reads++;
		} else if (op < r+a) {
		    if ((result = l.add(nums[val])) > 0) {
			successes++;
			failures += (result - 1);
		    } else failures -= (result + 1); // negative value
		    adds++;
		} else {
		    if ((result = l.remove(nums[val])) > 0) {
			successes++;
			failures += (result - 1);
		    } else failures -= (result + 1); // negative value
		    deletes++;
		}
	    }
	    
	    ctr.addReads(reads);
	    ctr.addAdds(adds);
	    ctr.addDeletes(deletes);
	    ctr.addSuccesses(successes);
	    ctr.addFailures(failures);
	}
    }
}


/* EOF */
